from __future__ import annotations

from typing import Final

DEFAULT_HANDSHAKE_TIMEOUT: Final[int] = 60
DEFAULT_SEND_TIMEOUT: Final[int] = 20
MAX_CONNECT_TIMEOUT: Final[int] = 60
TIMEOUT_INCREASE_FACTOR: Final[float] = 1.2

# CID to types mapping for Description Parser
DESCRIPTION_TYPES: Final = {
    1: "boolean",
    2: "integer",
    3: "enumeration",
    4: "float",
    5: "string",
    6: "dateTime",
    7: "temperatureCelsius",
    8: "temperatureFahrenheit",
    10: "hexBinary",
    16: "timeSpan",
    17: "percent",
    18: "dbm",
    19: "weight",
    20: "liquidVolume",
    21: "uidValue",
    22: "date",
    23: "time",
    24: "waterHardness",
    25: "point2D",
    26: "pose2D",
    27: "line2D",
    30: "rgb",
    31: "rpm",
    32: "flowRate",
    33: "length",
    34: "area",
    35: "power",
    36: "energy",
    37: "bigInteger",
    38: "identifier",
    39: "speed",
    40: "programInstruction",
    41: "weightPound",
    42: "localeString",
    43: "teaspoon",
    44: "tablespoon",
    45: "cup",
    46: "piece",
    47: "byteLength",
    48: "uuid",
    49: "timezone",
    50: "csv",
    51: "leaf",
    52: "bunch",
    53: "casket",
    54: "pinch",
    55: "stalk",
    56: "stick",
    57: "branch",
    58: "tray",
    59: "portion",
    61: "utcDateTime",
    62: "programRunSummary",
    63: "programSessionSummary",
    64: "liquidVolumeThroughput",
    65: "weightOunces",
    129: "booleanList",
    130: "integerList",
    131: "enumerationList",
    132: "floatList",
    133: "stringList",
    134: "dateTimeList",
    135: "temperatureCelsiusList",
    136: "temperatureFahrenheitList",
    138: "hexBinaryList",
    144: "timeSpanList",
    145: "percentList",
    146: "dBmList",
    147: "weightList",
    148: "liquidVolumeList",
    149: "uidValueList",
    150: "dateList",
    151: "timeList",
    152: "waterHardnessList",
    153: "point2DList",
    154: "pose2DList",
    155: "line2DList",
    156: "path",
    157: "polygon",
    158: "rgbList",
    159: "rpmList",
    160: "flowRateList",
    161: "lengthList",
    162: "areaList",
    163: "powerList",
    164: "energyList",
    165: "bigIntegerList",
    166: "identifierList",
    167: "speedList",
    168: "programInstructionList",
    169: "weightPoundList",
    170: "localeStringList",
    171: "teaspoonList",
    172: "tablespoonList",
    173: "cupList",
    174: "pieceList",
    175: "byteLengthList",
    176: "uuidList",
    177: "timezoneList",
    178: "csvList",
    179: "leafList",
    180: "bunchList",
    181: "casketList",
    182: "pinchList",
    183: "stalkList",
    184: "stickList",
    185: "branchList",
    186: "trayList",
    187: "portionList",
    189: "utcDateTimeList",
    190: "programRunSummaryList",
    191: "programSessionSummaryList",
    192: "liquidVolumeThroughputList",
    193: "weightOuncesList",
}

# CID to protocol type mapping for Description Parser
DESCRIPTION_PROTOCOL_TYPES: Final = {
    1: "Boolean",
    2: "Integer",
    3: "Integer",
    4: "Float",
    5: "String",
    6: "String",
    7: "Float",
    8: "Float",
    10: "String",
    16: "Integer",
    17: "Float",
    18: "Integer",
    19: "Integer",
    20: "Integer",
    21: "Integer",
    22: "String",
    23: "String",
    24: "Integer",
    25: "Object",
    26: "Object",
    27: "Object",
    30: "String",
    31: "Integer",
    32: "Integer",
    33: "Float",
    34: "Float",
    35: "Float",
    36: "Float",
    37: "Integer",
    38: "Integer",
    39: "Float",
    40: "Object",
    41: "Float",
    42: "Object",
    43: "Float",
    44: "Float",
    45: "Float",
    46: "Float",
    47: "Integer",
    48: "String",
    49: "Object",
    50: "String",
    51: "Float",
    52: "Float",
    53: "Float",
    54: "Float",
    55: "Float",
    56: "Float",
    57: "Float",
    58: "Integer",
    59: "Integer",
    61: "String",
    62: "Object",
    63: "Object",
    64: "Float",
    65: "Float",
    129: "Object",
    130: "Object",
    131: "Object",
    132: "Object",
    133: "Object",
    134: "Object",
    135: "Object",
    136: "Object",
    138: "Object",
    144: "Object",
    145: "Object",
    146: "Object",
    147: "Object",
    148: "Object",
    149: "Object",
    150: "Object",
    151: "Object",
    152: "Object",
    153: "Object",
    154: "Object",
    155: "Object",
    156: "Object",
    157: "Object",
    158: "Object",
    159: "Object",
    160: "Object",
    161: "Object",
    162: "Object",
    163: "Object",
    164: "Object",
    165: "Object",
    166: "Object",
    167: "Object",
    168: "Object",
    169: "Object",
    170: "Object",
    171: "Object",
    172: "Object",
    173: "Object",
    174: "Object",
    175: "Object",
    176: "Object",
    177: "Object",
    178: "Object",
    179: "Object",
    180: "Object",
    181: "Object",
    182: "Object",
    183: "Object",
    184: "Object",
    185: "Object",
    186: "Object",
    187: "Object",
    189: "Object",
    190: "Object",
    191: "Object",
    192: "Object",
    193: "Object",
}



from __future__ import annotations

import hmac
import logging
import ssl
from abc import abstractmethod
from base64 import urlsafe_b64decode

import aiohttp
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes


class HCSocket:
    """Socket Base class."""

    _URL_FORMAT = "ws://{host}:80/homeconnect"
    _session: aiohttp.ClientSession
    _websocket: aiohttp.ClientWebSocketResponse | None = None
    _owned_session: bool = False

    def __init__(
        self,
        host: str,
        session: aiohttp.ClientSession | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """
        Initialize.

        Args:
        ----
        host (str): Host
        session (Optional[aiohttp.ClientSession]): ClientSession
        logger (Optional[Logger]): Logger

        """
        if ":" in host:
            # is ipv6 address
            host = f"[{host}]"
        self._url = self._URL_FORMAT.format(host=host)
        if session is None:
            session = aiohttp.ClientSession()
            self._owned_session = True
        self._session = session
        if logger is None:
            self._logger = logging.getLogger(__name__)
        else:
            self._logger = logger.getChild("socket")

    @abstractmethod
    async def connect(self) -> None:
        """Connect to websocket."""
        self._logger.debug("Socket connecting to %s, mode=NONE", self._url)
        self._websocket = await self._session.ws_connect(self._url, heartbeat=20)

    @abstractmethod
    async def send(self, message: str) -> None:
        """Send message."""
        self._logger.debug("Send     %s: %s", self._url, message)
        await self._websocket.send_str(message)

    @abstractmethod
    async def _receive(self, message: aiohttp.WSMessage) -> str:
        """Recive message."""
        self._logger.debug("Received %s: %s", self._url, str(message.data))
        return str(message.data)

    async def close(self) -> None:
        """Close websocket."""
        self._logger.debug("Closing socket %s", self._url)
        if self._websocket:
            await self._websocket.close()
        if self._owned_session:
            await self._session.close()

    @property
    def closed(self) -> bool:
        """True if underlying websocket is closed."""
        if self._websocket:
            return self._websocket.closed
        return True

    def __aiter__(self) -> HCSocket:
        return self

    async def __anext__(self) -> str:
        msg = await self._websocket.__anext__()
        return await self._receive(msg)


class TlsSocket(HCSocket):
    """TLS (wss) Socket."""

    _URL_FORMAT = "wss://{host}:443/homeconnect"
    _ssl_context: ssl.SSLContext

    def __init__(
        self,
        host: str,
        psk64: str,
        session: aiohttp.ClientSession | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """
        TLS Socket.

        Args:
        ----
        host (str): Host
        psk64 (str): psk64 key
        session (Optional[aiohttp.ClientSession]): ClientSession
        logger (Optional[Logger]): Logger

        """
        # setup sslcontext
        psk = urlsafe_b64decode(psk64 + "===")
        self._ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        self._ssl_context.maximum_version = ssl.TLSVersion.TLSv1_2
        self._ssl_context.set_ciphers("PSK")
        self._ssl_context.check_hostname = False
        self._ssl_context.verify_mode = ssl.CERT_NONE
        self._ssl_context.set_psk_client_callback(lambda _: (None, psk))
        super().__init__(host, session, logger)

    async def connect(self) -> None:
        """Connect to websocket."""
        self._logger.debug("Socket connecting to %s, mode=TLS", self._url)
        self._websocket = await self._session.ws_connect(
            self._url, ssl=self._ssl_context, heartbeat=20
        )

    async def send(self, message: str) -> None:
        """Send message."""
        self._logger.debug("Send     %s: %s", self._url, message)
        await self._websocket.send_str(message)

    async def _receive(self, message: aiohttp.WSMessage) -> str:
        self._logger.debug("Received %s: %s", self._url, str(message.data))
        if message.type == aiohttp.WSMsgType.ERROR:
            raise message.data
        return str(message.data)


ENCRYPT_DIRECTION = b"\x45"  # 'E' in ASCII
DECRYPT_DIRECTION = b"\x43"  # 'C' in ASCII
MINIMUM_MESSAGE_LENGTH = 32


class AesSocket(HCSocket):
    """AES Socket."""

    _URL_FORMAT = "ws://{host}:80/homeconnect"
    _last_rx_hmac: bytes
    _last_tx_hmac: bytes

    def __init__(
        self,
        host: str,
        psk64: str,
        iv64: str,
        session: aiohttp.ClientSession | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """
        AES Socket.

        Args:
        ----
            host (str): Host
            psk64 (str): psk64 key
            iv64 (str): iv64
            session (Optional[aiohttp.ClientSession]): ClientSession
            logger (Optional[Logger]): Logger

        """
        psk = urlsafe_b64decode(psk64 + "===")
        self._iv = urlsafe_b64decode(iv64 + "===")
        self._enckey = hmac.digest(psk, b"ENC", digest="sha256")
        self._mackey = hmac.digest(psk, b"MAC", digest="sha256")

        super().__init__(host, session, logger)

    async def connect(self) -> None:
        """Connect to websocket."""
        self._last_rx_hmac = bytes(16)
        self._last_tx_hmac = bytes(16)

        self._aes_encrypt = AES.new(self._enckey, AES.MODE_CBC, self._iv)
        self._aes_decrypt = AES.new(self._enckey, AES.MODE_CBC, self._iv)

        self._logger.debug("Socket connecting to %s, mode=AES", self._url)
        self._websocket = await self._session.ws_connect(self._url, heartbeat=20)

    async def send(self, clear_msg: str) -> None:
        """Recive message."""
        self._logger.debug("Send     %s: %s", self._url, clear_msg)
        if isinstance(clear_msg, str):
            clear_msg = bytes(clear_msg, "utf-8")

        pad_len = 16 - (len(clear_msg) % 16)
        if pad_len == 1:
            pad_len += 16
        clear_msg = (
            clear_msg + b"\x00" + get_random_bytes(pad_len - 2) + bytearray([pad_len])
        )

        enc_msg = self._aes_encrypt.encrypt(clear_msg)

        hmac_msg = self._iv + ENCRYPT_DIRECTION + self._last_tx_hmac + enc_msg
        self._last_tx_hmac = hmac.digest(self._mackey, hmac_msg, digest="sha256")[0:16]

        await self._websocket.send_bytes(enc_msg + self._last_tx_hmac)

    async def _receive(self, message: aiohttp.WSMessage) -> str:
        if message.type != aiohttp.WSMsgType.BINARY:
            msg = "Message not of Type binary %s"
            msg.format(str(message))
            self._logger.warning(msg)
            raise ValueError(msg)

        buf = message.data
        if len(buf) < MINIMUM_MESSAGE_LENGTH:
            msg = "Message to short: %s"
            msg.format(str(message))
            self._logger.warning(msg)
            raise ValueError(msg)
        if len(buf) % 16 != 0:
            msg = "Unaligned Message %s"
            msg.format(str(message))
            self._logger.warning(msg)
            raise ValueError(msg)

        enc_msg = buf[0:-16]
        recv_hmac = buf[-16:]

        hmac_msg = self._iv + DECRYPT_DIRECTION + self._last_rx_hmac + enc_msg
        calculated_hmac = hmac.digest(self._mackey, hmac_msg, digest="sha256")[0:16]

        if not hmac.compare_digest(recv_hmac, calculated_hmac):
            msg = "HMAC Failure: %s"
            msg.format(str(message))
            self._logger.warning(msg)
            raise ValueError(msg)

        self._last_rx_hmac = recv_hmac

        msg = self._aes_decrypt.decrypt(enc_msg)
        pad_len = msg[-1]
        if len(msg) < pad_len:
            msg = "Padding Error %s"
            msg.format(str(message))
            self._logger.warning(msg)
            raise ValueError(msg)
        decoded_msg = msg[0:-pad_len].decode("utf-8")

        self._logger.debug("Received %s: %s", self._url, decoded_msg)
        return decoded_msg



from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock

import pytest
from homeconnect_websocket.entities import Access, Entity, EntityDescription, Status
from homeconnect_websocket.errors import AccessError
from homeconnect_websocket.message import Action, Message


def test_init_base() -> None:
    """Test Entity int ."""
    description = EntityDescription(uid=1, name="Test_Entity", protocolType="Integer")
    entity = Entity(description, AsyncMock())
    assert entity.uid == 1
    assert entity.name == "Test_Entity"
    assert entity.value is None
    assert entity.value_raw is None
    assert entity.enum is None


def test_init_full() -> None:
    """Test Entity init full."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=False,
        access=Access.READ,
        enumeration={"0": "Open", "1": "Closed"},
        protocolType="Integer",
    )
    entity = Entity(description, AsyncMock())
    assert entity.uid == 1
    assert entity.name == "Test_Entity"
    assert entity.value is None
    assert entity.value_raw is None
    assert entity.enum == {0: "Open", 1: "Closed"}
    assert entity._rev_enumeration == {"Open": 0, "Closed": 1}


def test_init_object() -> None:
    """Test Entity init full."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        default='{"test": "value", "other": 2}',
        available=False,
        access=Access.READ,
        protocolType="Object",
    )
    entity = Entity(description, AsyncMock())
    assert entity.uid == 1
    assert entity.name == "Test_Entity"
    assert entity.value == {"test": "value", "other": 2}
    assert entity.value_raw == {"test": "value", "other": 2}
    assert entity.enum is None


@pytest.mark.asyncio
async def test_update() -> None:
    """Test Entity.update()."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=False,
        access=Access.READ,
        protocolType="Integer",
    )
    entity = Entity(description, AsyncMock())
    await entity.update({"available": True, "access": Access.READ_WRITE, "value": 1})
    assert entity.value == 1
    assert entity.value_raw == 1


@pytest.mark.asyncio
async def test_update_enum() -> None:
    """Test Entity.update() with Enum."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=False,
        access=Access.READ,
        enumeration={"0": "Open", "1": "Closed"},
        protocolType="Integer",
    )
    entity = Entity(description, AsyncMock())
    await entity.update({"available": True, "access": Access.READ, "value": 1})
    assert entity.value == "Closed"
    assert entity.value_raw == 1


@pytest.mark.asyncio
async def test_set() -> None:
    """Test Entity.set_value()."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=True,
        access=Access.READ_WRITE,
        protocolType="Integer",
    )
    appliance = AsyncMock()
    entity = Entity(description, appliance)
    await entity.set_value(1)
    appliance.session.send_sync.assert_called_with(
        Message(
            resource="/ro/values",
            action=Action.POST,
            data={"uid": 1, "value": 1},
        )
    )


@pytest.mark.asyncio
async def test_set_raw() -> None:
    """Test Entity.set_value_raw()."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=True,
        access=Access.READ_WRITE,
        protocolType="Integer",
    )
    appliance = AsyncMock()
    entity = Entity(description, appliance)
    await entity.set_value_raw(1)
    appliance.session.send_sync.assert_called_with(
        Message(
            resource="/ro/values",
            action=Action.POST,
            data={"uid": 1, "value": 1},
        )
    )


@pytest.mark.asyncio
async def test_set_enum() -> None:
    """Test Entity.set_value() with Enum."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=True,
        access=Access.READ_WRITE,
        enumeration={"0": "Open", "1": "Closed"},
        protocolType="Integer",
    )
    appliance = AsyncMock()
    entity = Entity(description, appliance)
    await entity.set_value("Closed")
    appliance.session.send_sync.assert_called_with(
        Message(
            resource="/ro/values",
            action=Action.POST,
            data={"uid": 1, "value": 1},
        )
    )


@pytest.mark.asyncio
async def test_set_raw_enum() -> None:
    """Test Entity.set_value_raw() with Enum."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=True,
        access=Access.READ_WRITE,
        enumeration={0: "Open", 1: "Closed"},
        protocolType="Integer",
    )
    appliance = AsyncMock()
    entity = Entity(description, appliance)
    await entity.set_value_raw(0)
    appliance.session.send_sync.assert_called_once_with(
        Message(
            resource="/ro/values",
            action=Action.POST,
            data={"uid": 1, "value": 0},
        )
    )


@pytest.mark.asyncio
async def test_callback() -> None:
    """Test Entity callback."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=False,
        access=Access.READ,
        protocolType="Integer",
    )
    entity = Entity(description, AsyncMock())

    callback_1 = MagicMock()
    callback_2 = MagicMock()
    entity.register_callback(callback_1)
    entity.register_callback(callback_2)

    assert entity._callbacks == {callback_1, callback_2}

    await entity.update({"available": True, "access": Access.READ_WRITE, "value": 1})

    callback_1.assert_called_once_with(entity)
    callback_2.assert_called_once_with(entity)

    entity.unregister_callback(callback_1)
    entity.unregister_callback(callback_2)

    assert entity._callbacks == set()


@pytest.mark.asyncio
async def test_access() -> None:
    """Test Entity Access check."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=False,
        access=Access.READ,
        enumeration={"0": "Open", "1": "Closed"},
        protocolType="Integer",
    )
    entity = Status(description, AsyncMock())

    with pytest.raises(AccessError):
        await entity.set_value("Open")

    with pytest.raises(AccessError):
        await entity.set_value_raw(1)

    await entity.update({"access": Access.READ_WRITE})

    with pytest.raises(AccessError):
        await entity.set_value("Open")

    with pytest.raises(AccessError):
        await entity.set_value_raw(1)

    await entity.update({"available": True})

    await entity.set_value("Open")
    await entity.set_value_raw(1)


@pytest.mark.asyncio
async def test_dump() -> None:
    """Test Entity state dump."""
    description = EntityDescription(
        uid=1,
        name="Test_Entity",
        available=False,
        access=Access.READ,
        min=0,
        max=10,
        stepSize=2,
        enumeration={0: "a", 1: "b"},
        protocolType="Integer",
    )
    entity = Status(description, AsyncMock())
    await entity.update({"value": 1})
    assert entity.dump() == {
        "uid": 1,
        "name": "Test_Entity",
        "available": False,
        "value": "b",
        "value_raw": 1,
        "enum": {0: "a", 1: "b"},
        "access": "read",
        "min": 0,
        "max": 10,
        "step": 2,
    }



"""Parser to parse device description from XML-Files."""

from __future__ import annotations

from functools import partial
from typing import TextIO, TypedDict
from xml.parsers.expat import ExpatError

import xmltodict

from .const import DESCRIPTION_PROTOCOL_TYPES, DESCRIPTION_TYPES
from .entities import (
    DeviceDescription,
    EntityDescription,
    OptionDescription,
)
from .errors import ParserError
from .helpers import convert_bool


class FeatureMap(TypedDict):
    """Typing for Feature mapping."""

    feature: dict[int, str]
    error: dict[int, str]
    enumeration: dict[int, dict[int, str]]


def parse_feature_mapping(feature_mapping: dict) -> FeatureMap:
    """Parse Feature mapping."""
    features = FeatureMap(feature={}, error={}, enumeration={})

    try:
        for feature in feature_mapping["featureDescription"]["feature"]:
            features["feature"][int(feature["@refUID"], base=16)] = feature["#text"]
    except (KeyError, ValueError, TypeError) as exc:
        msg = "Error while parsing on 'featureDescription'"
        raise ParserError(msg) from exc

    try:
        for error in feature_mapping["errorDescription"]["error"]:
            features["error"][int(error["@refEID"], base=16)] = error["#text"]
    except (KeyError, ValueError, TypeError) as exc:
        msg = "Error while parsing on 'errorDescription'"
        raise ParserError(msg) from exc

    try:
        for enum in feature_mapping["enumDescriptionList"]["enumDescription"]:
            temp_enum = {}
            for key in enum["enumMember"]:
                temp_enum[int(key["@refValue"])] = key["#text"]
            features["enumeration"][int(enum["@refENID"], base=16)] = temp_enum
    except (KeyError, ValueError, TypeError) as exc:
        msg = "Error while parsing on 'enumDescriptionList'"
        raise ParserError(msg) from exc

    return features


def parse_options(element: list[dict] | dict) -> list[OptionDescription]:
    """Parse Programs Options."""
    options = []
    for option in element:
        option_out = {}
        if "@access" in option:
            option_out["access"] = option["@access"].lower()
        if "@available" in option:
            option_out["available"] = convert_bool(option["@available"])
        if "@liveUpdate" in option:
            option_out["liveUpdate"] = convert_bool(option["@liveUpdate"])
        if "@refUID" in option:
            option_out["refUID"] = int(option["@refUID"], base=16)
        if "@default" in option:
            option_out["default"] = option["@default"]
        options.append(option_out)
    return options


def parse_element(
    description: DeviceDescription,
    xml_description: dict,
    features: FeatureMap,
    key: str,
    *,
    is_list: bool = True,
) -> None:
    """Parse Element."""
    element_out = EntityDescription()

    for attr_name, attr_value in xml_description.items():
        try:
            if attr_name == "@uid":
                element_out["uid"] = int(attr_value, base=16)
                element_out["name"] = features["feature"][
                    int(xml_description["@uid"], base=16)
                ]
            elif attr_name == "@refCID":
                element_out["contentType"] = DESCRIPTION_TYPES[int(attr_value, base=16)]
                element_out["protocolType"] = DESCRIPTION_PROTOCOL_TYPES[
                    int(attr_value, base=16)
                ]
                element_out["refCID"] = int(attr_value, base=16)
            elif attr_name == "@enumerationType":
                element_out["enumeration"] = features["enumeration"][
                    int(attr_value, base=16)
                ]
            elif attr_name in (
                "@available",
                "@notifyOnChange",
                "@passwordProtected",
                "@liveUpdate",
                "@fullOptionSet",
                "@validate",
            ):
                element_out[attr_name.strip("@")] = convert_bool(attr_value)
            elif attr_name in ("@access", "@execution"):
                element_out[attr_name.strip("@")] = attr_value.lower()
            elif attr_name == "option":
                element_out["options"] = parse_options(xml_description["option"])
            elif attr_name == "@refDID":
                element_out["refDID"] = int(attr_value, base=16)
            else:
                element_out[attr_name.strip("@")] = attr_value
        except (KeyError, ValueError, TypeError) as exc:
            msg = f"Error while parsing '{attr_name}' in '{key}'"
            raise ParserError(msg) from exc

    if is_list:
        description[key].append(element_out)
    else:
        description[key] = element_out


def parse_elements(
    description: DeviceDescription, xml_description: list[dict], features: FeatureMap
) -> None:
    """Parse list of Element."""
    for element, parser in PARSERS.items():
        if element in xml_description:
            description_elements = xml_description[element]
            if not isinstance(description_elements, list):
                description_elements = [description_elements]
            for description_element in description_elements:
                parser["parser"](description, description_element, features)


def parse_info(
    description: DeviceDescription,
    xml_description: dict,
    features: FeatureMap,  # noqa: ARG001
    key: str,
) -> None:
    """Parse Device Info."""
    try:
        description[key] = {
            "brand": xml_description["brand"],
            "type": xml_description["type"],
            "model": xml_description["model"],
            "version": int(xml_description["version"]),
            "revision": int(xml_description["revision"]),
        }
    except (KeyError, ValueError, TypeError) as exc:
        msg = "Error while parsing 'Device Info'"
        raise ParserError(msg) from exc


PARSERS = {
    "description": {"parser": partial(parse_info, key="info")},
    "option": {"parser": partial(parse_element, key="option")},
    "optionList": {"parser": parse_elements},
    "status": {"parser": partial(parse_element, key="status")},
    "statusList": {"parser": parse_elements, "key": "status"},
    "setting": {"parser": partial(parse_element, key="setting")},
    "settingList": {"parser": parse_elements, "key": "setting"},
    "event": {"parser": partial(parse_element, key="event")},
    "eventList": {"parser": parse_elements},
    "command": {"parser": partial(parse_element, key="command")},
    "commandList": {"parser": parse_elements},
    "program": {"parser": partial(parse_element, key="program")},
    "programGroup": {"parser": parse_elements},
    "activeProgram": {
        "parser": partial(parse_element, key="activeProgram", is_list=False)
    },
    "selectedProgram": {
        "parser": partial(parse_element, key="selectedProgram", is_list=False)
    },
    "protectionPort": {
        "parser": partial(parse_element, key="protectionPort", is_list=False)
    },
}


def parse_device_description(
    device_description_xml: str | TextIO, feature_mapping_xml: str | TextIO
) -> DeviceDescription:
    """
    Parse device description from XML-Files.

    Args:
    ----
        device_description_xml (str | TextIO): Device description XML-File
        feature_mapping_xml (str | TextIO): Feature mapping XML-File

    """
    try:
        xml_description = xmltodict.parse(
            device_description_xml,
            force_list=(
                "option",
                "optionList",
                "status",
                "statusList",
                "setting",
                "settingList",
                "event",
                "eventList",
                "command",
                "commandList",
                "program",
                "programGroup",
            ),
        )["device"]
    except ExpatError as exc:
        msg = "Error while parsing Device Description XML-File"
        raise ParserError(msg) from exc

    try:
        feature_mapping = xmltodict.parse(
            feature_mapping_xml,
            force_list=("feature", "error", "enumDescription", "enumMember"),
        )["featureMappingFile"]
    except ExpatError as exc:
        msg = "Error while parsing Feature Mapping XML-File"
        raise ParserError(msg) from exc

    features = parse_feature_mapping(feature_mapping)

    description = DeviceDescription(
        status=[], option=[], setting=[], event=[], command=[], program=[]
    )
    parse_elements(description, xml_description, features)

    return description


def main() -> None:
    """For CLI Parser."""
    import json
    from argparse import ArgumentParser
    from pathlib import Path

    arg_parser = ArgumentParser(
        description="HomeConnect Websocket Description Parser",
        usage="%(prog)s -d DeviceDescription.xml -f FeatureMapping.xml -o output.json",
    )
    arg_parser.add_argument(
        "-d",
        type=Path,
        required=True,
        dest="description_path",
        help="Device description files",
    )
    arg_parser.add_argument(
        "-f",
        type=Path,
        required=True,
        dest="feature_path",
        help="Feature mapping files",
    )
    arg_parser.add_argument(
        "-o", type=Path, required=True, dest="output_file", help="Output file"
    )
    args = arg_parser.parse_args()

    with Path(args.description_path).open() as file:
        description_file = file.read()

    with Path(args.feature_path).open() as file:
        feature_file = file.read()

    description = parse_device_description(description_file, feature_file)

    with Path(args.output_file).open("w") as file:
        json.dump(description, file, indent=4)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
)
from unittest.mock import ANY, AsyncMock

import aiohttp
import pytest
from homeconnect_websocket.message import Action, Message
from homeconnect_websocket.session import HCSession
from homeconnect_websocket.testutils import TEST_APP_ID, TEST_APP_NAME

from const import (
    CLIENT_MESSAGE_ID,
    SERVER_MESSAGE_ID,
    SESSION_ID,
)

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from tests.utils import ApplianceServerAes

    from utils import ApplianceServer


@pytest.mark.asyncio
async def test_session_connect_tls(
    appliance_server_tls: Callable[..., Awaitable[ApplianceServer]],
) -> None:
    """Test Session connection."""
    appliance_server = await appliance_server_tls()
    session = HCSession(
        appliance_server.host,
        app_name=TEST_APP_NAME,
        app_id=TEST_APP_ID,
        psk64=appliance_server.psk64,
    )
    session.handshake = False
    message_handler = AsyncMock()

    assert not session.connected
    await session.connect(message_handler)
    assert session.connected

    await session.close()
    assert not session.connected

    message_handler.assert_called_once_with(
        Message(
            sid=SESSION_ID,
            msg_id=SERVER_MESSAGE_ID,
            resource="/ei/initialValues",
            version=2,
            action=Action.POST,
            data=[{"edMsgID": CLIENT_MESSAGE_ID}],
            code=None,
        )
    )


@pytest.mark.asyncio
async def test_session_connect_aes(
    appliance_server_aes: Callable[..., Awaitable[ApplianceServerAes]],
) -> None:
    """Test Session connection failing."""
    appliance_server = await appliance_server_aes()
    session = HCSession(
        appliance_server.host,
        app_name=TEST_APP_NAME,
        app_id=TEST_APP_ID,
        psk64=appliance_server.psk64,
        iv64=appliance_server.iv64,
    )
    session.handshake = False
    message_handler = AsyncMock()

    assert not session.connected
    await session.connect(message_handler)
    assert session.connected

    await session.close()
    assert not session.connected

    message_handler.assert_called_once_with(
        Message(
            sid=SESSION_ID,
            msg_id=SERVER_MESSAGE_ID,
            resource="/ei/initialValues",
            version=2,
            action=Action.POST,
            data=[{"edMsgID": CLIENT_MESSAGE_ID}],
            code=None,
        )
    )


@pytest.mark.asyncio
async def test_session_handshake(
    appliance_server: Callable[..., Awaitable[ApplianceServer]],
) -> None:
    """Test Session Handshake."""
    appliance = await appliance_server()
    session = HCSession(
        appliance.host,
        app_name=TEST_APP_NAME,
        app_id=TEST_APP_ID,
        psk64=None,
    )
    message_handler = AsyncMock()
    await session.connect(message_handler)
    await session.close()

    assert appliance.messages[0] == Message(
        sid=10,
        msg_id=20,
        resource="/ei/initialValues",
        version=2,
        action=Action.RESPONSE,
        data=[
            {
                "deviceType": "Application",
                "deviceName": "Test Device",
                "deviceID": "c6683b15",
            }
        ],
    )

    assert appliance.messages[1] == Message(
        sid=10, msg_id=30, resource="/ci/services", version=1, action=Action.GET
    )

    assert appliance.messages[2] == Message(
        sid=10,
        msg_id=31,
        resource="/ci/authentication",
        version=3,
        action=Action.GET,
        data=[{"nonce": ANY}],
    )

    assert appliance.messages[3] == Message(
        sid=10, msg_id=32, resource="/ci/info", version=3, action=Action.GET
    )

    assert appliance.messages[4] == Message(
        sid=10, msg_id=33, resource="/iz/info", version=1, action=Action.GET
    )

    assert appliance.messages[5] == Message(
        sid=10, msg_id=34, resource="/ei/deviceReady", version=2, action=Action.NOTIFY
    )

    assert appliance.messages[6] == Message(
        sid=10, msg_id=35, resource="/ni/info", version=1, action=Action.GET
    )
    assert appliance.messages[7] == Message(
        sid=10,
        msg_id=36,
        resource="/ro/allMandatoryValues",
        version=1,
        action=Action.GET,
    )

    assert appliance.messages[8] == Message(
        sid=10, msg_id=37, resource="/ro/values", version=1, action=Action.GET
    )

    assert appliance.messages[9] == Message(
        sid=10,
        msg_id=38,
        resource="/ro/allDescriptionChanges",
        version=1,
        action=Action.GET,
    )


@pytest.mark.asyncio
async def test_session_connect_failed() -> None:
    """Test Session connction failing."""
    session = HCSession(
        "127.0.0.1",
        app_name=TEST_APP_NAME,
        app_id=TEST_APP_ID,
        psk64=None,
    )
    with pytest.raises(aiohttp.ClientConnectionError):
        await session.connect(AsyncMock())
    assert not session.connected



from __future__ import annotations

import json
from base64 import urlsafe_b64encode
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock

import pytest
from aiohttp import WSMessage, WSMsgType
from Crypto.Random import get_random_bytes
from homeconnect_websocket.hc_socket import AesSocket, TlsSocket
from homeconnect_websocket.testutils import TEST_IV64, TEST_PSK64

from const import CLIENT_MESSAGE_ID, SERVER_MESSAGE_ID, SESSION_ID
from utils import AesServerEncryption, ApplianceServer, ApplianceServerAes

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


@pytest.mark.asyncio
async def test_connect_tls(
    appliance_server_tls: Callable[..., Awaitable[ApplianceServer]],
) -> None:
    """Test TLS Socket connect."""
    appliance_server = await appliance_server_tls()
    socket = TlsSocket(appliance_server.host, psk64=appliance_server.psk64)
    assert socket.closed
    await socket.connect()

    assert not socket.closed
    async for msg in socket:
        assert json.loads(msg) == {
            "sID": SESSION_ID,
            "msgID": SERVER_MESSAGE_ID,
            "resource": "/ei/initialValues",
            "version": 2,
            "action": "POST",
            "data": [{"edMsgID": CLIENT_MESSAGE_ID}],
        }
        break
    assert not socket.closed

    await socket.send("Pong")
    await socket.close()
    assert socket.closed

    assert appliance_server.messages[0] == "Pong"


@pytest.mark.asyncio
async def test_connect_aes(
    appliance_server_aes: Callable[..., Awaitable[ApplianceServerAes]],
) -> None:
    """Test AES Socket connect."""
    appliance_server = await appliance_server_aes()
    socket = AesSocket(
        host=appliance_server.host,
        psk64=appliance_server.psk64,
        iv64=appliance_server.iv64,
    )
    assert socket.closed
    await socket.connect()

    assert not socket.closed
    async for msg in socket:
        assert json.loads(msg) == {
            "sID": SESSION_ID,
            "msgID": SERVER_MESSAGE_ID,
            "resource": "/ei/initialValues",
            "version": 2,
            "action": "POST",
            "data": [{"edMsgID": CLIENT_MESSAGE_ID}],
        }
        break
    assert not socket.closed

    await socket.send("Pong")
    await socket.close()
    assert socket.closed

    assert appliance_server.messages[0] == "Pong"


@pytest.mark.asyncio
async def test_ase_padding(
    appliance_server_aes: Callable[..., Awaitable[ApplianceServerAes]],
) -> None:
    """Test AES Socket padding."""
    appliance_server = await appliance_server_aes()
    socket = AesSocket(
        host=appliance_server.host,
        psk64=appliance_server.psk64,
        iv64=appliance_server.iv64,
    )
    await socket.connect()

    messages = [  # len/padding
        "",  # 0/16
        "a",  # 1/15
        "ab",  # 2/14
        "abcdefghijklmno",  # 15/17
        "abcdefghijklmnop",  # 16/16
    ]
    for msg in messages:
        await socket.send(msg)
    await socket.close()

    assert appliance_server.messages == messages


@pytest.mark.asyncio
async def test_ase_padding_error() -> None:
    """Test AES Socket with padding error."""
    encryption = AesServerEncryption(psk64=TEST_PSK64, iv64=TEST_IV64)

    socket = AesSocket("", psk64=TEST_PSK64, iv64=TEST_IV64)
    socket._session = AsyncMock()
    await socket.connect()

    clear_msg = get_random_bytes(32)
    pad_len = 16 - (len(clear_msg) % 16)
    if pad_len == 1:
        pad_len += 16
    clear_msg = clear_msg + b"\x00" + get_random_bytes(pad_len - 2) + bytearray([64])

    enc_msg = encryption.aes_encrypt.encrypt(clear_msg)
    enc_msg += encryption.hmac_encrypt(enc_msg)

    msg = WSMessage(type=WSMsgType.BINARY, data=enc_msg, extra=None)
    with pytest.raises(ValueError, match="Padding Error"):
        await socket._receive(msg)


@pytest.mark.asyncio
async def test_ase_wrong_msg_type() -> None:
    """Test AES Socket with Message not of type binary."""
    psk64 = urlsafe_b64encode(get_random_bytes(32)).decode()
    iv64 = urlsafe_b64encode(get_random_bytes(16)).decode()
    socket = AesSocket("", psk64=psk64, iv64=iv64)

    msg = WSMessage(type=WSMsgType.PING, data=None, extra=None)
    with pytest.raises(ValueError, match="Message not of Type binary"):
        await socket._receive(msg)


@pytest.mark.asyncio
async def test_ase_msg_to_short() -> None:
    """Test AES Socket with Message to short."""
    encryption = AesServerEncryption(psk64=TEST_PSK64, iv64=TEST_IV64)
    encryption.reset()

    msg_data = encryption.encrypt(get_random_bytes(2))

    socket = AesSocket("", psk64=TEST_PSK64, iv64=TEST_IV64)
    socket._session = AsyncMock()
    await socket.connect()

    msg = WSMessage(type=WSMsgType.BINARY, data=msg_data[:-16], extra=None)
    with pytest.raises(ValueError, match="Message to short"):
        await socket._receive(msg)


@pytest.mark.asyncio
async def test_ase_msg_unaligned() -> None:
    """Test AES Socket with unaligned Message."""
    encryption = AesServerEncryption(psk64=TEST_PSK64, iv64=TEST_IV64)
    encryption.reset()

    msg_data = encryption.encrypt(get_random_bytes(32))
    msg = WSMessage(type=WSMsgType.BINARY, data=msg_data[:-1], extra=None)

    socket = AesSocket("", psk64=TEST_PSK64, iv64=TEST_IV64)
    socket._session = AsyncMock()
    await socket.connect()

    with pytest.raises(ValueError, match="Unaligned Message"):
        await socket._receive(msg)


@pytest.mark.asyncio
async def test_ase_hmac_failure() -> None:
    """Test AES Socket with HMAC failure."""
    encryption = AesServerEncryption(psk64=TEST_PSK64, iv64=TEST_IV64)
    encryption.reset()

    socket = AesSocket("", psk64=TEST_PSK64, iv64=TEST_IV64)
    socket._session = AsyncMock()
    await socket.connect()

    msg_data = encryption.encrypt(get_random_bytes(32))
    msg_data = msg_data[:-16] + get_random_bytes(16)

    msg = WSMessage(type=WSMsgType.BINARY, data=msg_data, extra=None)
    with pytest.raises(ValueError, match="HMAC Failure"):
        await socket._receive(msg)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from homeconnect_websocket.entities import (
    EntityDescription,
    OptionDescription,
    Program,
)
from homeconnect_websocket.message import Action, Message

if TYPE_CHECKING:
    from homeconnect_websocket.testutils import MockApplianceType


@pytest.mark.asyncio
async def test_init(
    mock_homeconnect_appliance: MockApplianceType,
) -> None:
    """Test Progrmm Entity init."""
    description = EntityDescription(
        uid=1,
        name="Test_Program",
        options=[
            OptionDescription(refUID=10000, protocolType="Integer"),
            OptionDescription(refUID=10001, protocolType="Integer"),
            OptionDescription(refUID=10002, protocolType="Integer"),
        ],
    )
    appliance = await mock_homeconnect_appliance()
    entity = Program(description, appliance)

    assert entity._options == [
        appliance.entities_uid[10000],
        appliance.entities_uid[10001],
        appliance.entities_uid[10002],
    ]


@pytest.mark.asyncio
async def test_init_no_options(
    mock_homeconnect_appliance: MockApplianceType,
) -> None:
    """Test Progrmm Entity init."""
    description = EntityDescription(uid=1, name="Test_Program")

    appliance = await mock_homeconnect_appliance()
    entity = Program(description, appliance)

    assert entity._options == []


@pytest.mark.asyncio
async def test_select(
    mock_homeconnect_appliance: MockApplianceType,
) -> None:
    """Test select Progrmm."""
    description = EntityDescription(
        uid=1,
        name="Test_Program",
        options=[
            OptionDescription(refUID=10000),
            OptionDescription(refUID=10001),
            OptionDescription(refUID=10002),
        ],
    )
    appliance = await mock_homeconnect_appliance()
    entity = Program(description, appliance)

    await entity.select()
    appliance.session.send_sync.assert_called_once_with(
        Message(
            resource="/ro/selectedProgram",
            action=Action.POST,
            data={"program": 1, "options": []},
        )
    )


@pytest.mark.asyncio
async def test_start(
    mock_homeconnect_appliance: MockApplianceType,
) -> None:
    """Test start Progrmm."""
    description = EntityDescription(
        uid=1,
        name="Test_Program",
        options=[
            OptionDescription(refUID=10000),
            OptionDescription(refUID=10001),
            OptionDescription(refUID=10002),
        ],
    )
    appliance = await mock_homeconnect_appliance()
    entity = Program(description, appliance)

    await entity.start()
    appliance.session.send_sync.assert_called_once_with(
        Message(
            resource="/ro/activeProgram",
            action=Action.POST,
            data={
                "program": 1,
                "options": [
                    {"uid": 10000, "value": True},
                    {"uid": 10001, "value": "str"},
                ],
            },
        )
    )



"""Testing Utilities."""

from __future__ import annotations

from typing import Protocol
from unittest.mock import AsyncMock, Mock

import pytest

from homeconnect_websocket import HomeAppliance
from homeconnect_websocket.session import HCSession

from .entities import DeviceDescription

TEST_PSK64 = "whZJhkPa3a1hkuDdI3twHdqi1qhTxjnKE8954_zyY_E="
TEST_IV64 = "ofi7M1WB98sJeM2H1Ew3XA=="

TEST_APP_ID = "c6683b15"
TEST_APP_NAME = "Test Device"

BASE_DESCRIPTION = DeviceDescription(
    info={}, status=[], setting=[], event=[], command=[], option=[], program=[]
)


class MockAppliance(HomeAppliance):
    """Mock HomeAppliance."""

    def __init__(
        self,
        description: DeviceDescription,
        host: str,
        app_name: str,
        app_id: str,
        psk64: str,
        iv64: str | None = None,
    ) -> None:
        """
        HomeConnect Appliance.

        Args:
        ----
            description (DeviceDescription): parsed Device description
            host (str): Host
            app_name (str): Name used to identify this App
            app_id (str): ID used to identify this App
            psk64 (str): urlsafe base64 encoded psk key
            iv64 (Optional[str]): urlsafe base64 encoded iv64 key (only AES)

        """
        session_mock = Mock(return_value=AsyncMock(spec=HCSession))
        self.session = session_mock(host, app_name, app_id, psk64, iv64)
        _description = BASE_DESCRIPTION.copy()
        _description.update(description)
        self.info = _description.get("info", {})

        self.entities_uid = {}
        self.entities = {}
        self.status = {}
        self.settings = {}
        self.events = {}
        self.commands = {}
        self.options = {}
        self.programs = {}
        self._create_entities(_description)


@pytest.fixture
def mock_homeconnect_appliance() -> MockApplianceType:
    """Mock HomeAppliance for testing."""

    async def go(
        description: DeviceDescription = DESCRIPTION,
        host: str = "127.0.0.1",
        app_name: str = TEST_APP_NAME,
        app_id: str = TEST_APP_ID,
        psk64: str = TEST_PSK64,
        iv64: str | None = TEST_IV64,
    ) -> MockAppliance:
        return MockAppliance(description, host, app_name, app_id, psk64, iv64)

    return go


DESCRIPTION = DeviceDescription(
    {
        "info": {
            "brand": "HomeConnectWS",
            "type": "Dishwasher",
            "model": "FFAA1234",
            "version": 5,
            "revision": 0,
        },
        "status": [
            {
                "access": "read",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 5,
                "name": "BSH.Common.Status.BackendConnected",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 517,
                "name": "BSH.Common.Status.RemoteControlStartAllowed",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 523,
                "name": "BSH.Common.Status.RemoteControlActive",
            },
            {
                "access": "read",
                "available": True,
                "enumeration": {"0": "Open", "1": "Closed"},
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 527,
                "name": "BSH.Common.Status.DoorState",
            },
            {
                "access": "read",
                "available": True,
                "enumeration": {
                    "0": "Inactive",
                    "1": "Ready",
                    "2": "DelayedStart",
                    "3": "Run",
                    "4": "Pause",
                    "5": "ActionRequired",
                    "6": "Finished",
                    "7": "Error",
                    "8": "Aborting",
                },
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 552,
                "name": "BSH.Common.Status.OperationState",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "identifier",
                "protocolType": "Integer",
                "uid": 592,
                "name": "BSH.Common.Status.SoftwareUpdateTransactionID",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "stringList",
                "protocolType": "Object",
                "uid": 614,
                "name": "BSH.Common.Status.ErrorCodesList",
            },
            {
                "access": "read",
                "available": False,
                "contentType": "integer",
                "protocolType": "Integer",
                "uid": 615,
                "name": "BSH.Common.Status.Program.All.Count.Started",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "programSessionSummary",
                "protocolType": "Object",
                "uid": 625,
                "name": "BSH.Common.Status.ProgramSessionSummary.Latest",
            },
            {
                "access": "read",
                "available": True,
                "enumeration": {
                    "0": "ProgramFinished",
                    "1": "ProgramAbortedByUser",
                    "2": "ProgramAbortedByAppliance",
                    "3": "ProgramAbortedByApplianceCriticalError",
                },
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 626,
                "name": "BSH.Common.Status.ProgramRunDetail.EndTrigger",
            },
            {
                "access": "read",
                "available": True,
                "enumeration": {"0": "AsList", "1": "AsButtons"},
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 32771,
                "name": "BSH.Common.Status.Favorite.Handling",
            },
        ],
        "setting": [
            {
                "access": "readwrite",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 3,
                "name": "BSH.Common.Setting.AllowBackendConnection",
            },
            {
                "access": "readwrite",
                "available": True,
                "enumeration": {"1": "Off", "2": "On"},
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 539,
                "name": "BSH.Common.Setting.PowerState",
            },
            {
                "access": "readwrite",
                "available": True,
                "enumeration": {
                    "0": "Monitoring",
                    "1": "ManualRemoteStart",
                    "2": "PermanentRemoteStart",
                },
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 15,
                "name": "BSH.Common.Setting.RemoteControlLevel",
            },
            {
                "access": "readwrite",
                "available": True,
                "enumeration": {"0": "Off", "1": "Program"},
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 32824,
                "name": "BSH.Common.Setting.Favorite.001.Functionality",
            },
            {
                "access": "readwrite",
                "available": True,
                "max": "30",
                "min": "0",
                "contentType": "string",
                "protocolType": "String",
                "uid": 32825,
                "name": "BSH.Common.Setting.Favorite.001.Name",
            },
            {
                "access": "readwrite",
                "available": True,
                "max": "1",
                "min": "0",
                "contentType": "programInstructionList",
                "protocolType": "Object",
                "uid": 32826,
                "name": "BSH.Common.Setting.Favorite.001.Program",
            },
        ],
        "event": [
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "acknowledge",
                "level": "hint",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 21,
                "name": "BSH.Common.Event.SoftwareUpdateAvailable",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "none",
                "level": "hint",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 46,
                "name": "BSH.Common.Event.ConfirmPermanentRemoteStart",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "none",
                "level": "critical",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 525,
                "name": "BSH.Common.Event.AquaStopOccured",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "none",
                "level": "hint",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 540,
                "name": "BSH.Common.Event.ProgramFinished",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "none",
                "level": "alert",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 543,
                "name": "BSH.Common.Event.LowWaterPressure",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "acknowledge",
                "level": "hint",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 545,
                "name": "BSH.Common.Event.ProgramAborted",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "none",
                "level": "warning",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 577,
                "name": "BSH.Common.Event.ConnectLocalWiFi",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "acknowledge",
                "level": "hint",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 593,
                "name": "BSH.Common.Event.SoftwareDownloadAvailable",
            },
            {
                "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
                "handling": "acknowledge",
                "level": "hint",
                "contentType": "enumeration",
                "protocolType": "Integer",
                "uid": 595,
                "name": "BSH.Common.Event.SoftwareUpdateSuccessful",
            },
        ],
        "command": [
            {
                "access": "writeonly",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 1,
                "name": "BSH.Common.Command.DeactivateWiFi",
            },
            {
                "access": "writeonly",
                "available": True,
                "contentType": "uidValue",
                "protocolType": "Integer",
                "uid": 6,
                "name": "BSH.Common.Command.AcknowledgeEvent",
            },
            {
                "access": "writeonly",
                "available": True,
                "contentType": "uidValue",
                "protocolType": "Integer",
                "uid": 16,
                "name": "BSH.Common.Command.RejectEvent",
            },
            {
                "access": "none",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 512,
                "name": "BSH.Common.Command.AbortProgram",
            },
            {
                "access": "writeonly",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 553,
                "name": "BSH.Common.Command.ApplyFactoryReset",
            },
            {
                "access": "writeonly",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 555,
                "name": "BSH.Common.Command.DeactivateRemoteControlStart",
            },
            {
                "access": "writeonly",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 611,
                "name": "BSH.Common.Command.AllowSoftwareUpdate",
            },
            {
                "access": "writeonly",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 594,
                "name": "BSH.Common.Command.AllowSoftwareDownload",
            },
            {
                "access": "writeonly",
                "available": True,
                "contentType": "waterHardness",
                "protocolType": "Integer",
                "uid": 556,
                "name": "BSH.Common.Command.SetWaterHardness",
            },
        ],
        "option": [
            {
                "access": "read",
                "available": True,
                "contentType": "percent",
                "protocolType": "Float",
                "uid": 542,
                "name": "BSH.Common.Option.ProgramProgress",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "timeSpan",
                "protocolType": "Integer",
                "uid": 544,
                "name": "BSH.Common.Option.RemainingProgramTime",
            },
            {
                "access": "read",
                "available": True,
                "initValue": "1",
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 549,
                "name": "BSH.Common.Option.RemainingProgramTimeIsEstimated",
            },
            {
                "access": "read",
                "available": True,
                "max": "86400",
                "min": "0",
                "contentType": "timeSpan",
                "protocolType": "Integer",
                "uid": 558,
                "name": "BSH.Common.Option.StartInRelative",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "percent",
                "protocolType": "Float",
                "uid": 561,
                "name": "BSH.Common.Option.EnergyForecast",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "percent",
                "protocolType": "Float",
                "uid": 562,
                "name": "BSH.Common.Option.WaterForecast",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "string",
                "protocolType": "String",
                "uid": 32772,
                "name": "BSH.Common.Option.ProgramName",
            },
            {
                "access": "readwrite",
                "available": True,
                "contentType": "uidValue",
                "protocolType": "Integer",
                "uid": 32773,
                "name": "BSH.Common.Option.BaseProgram",
            },
            {
                "access": "readwrite",
                "available": True,
                "initValue": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 10000,
                "name": "Test_Option.1",
            },
            {
                "access": "readwrite",
                "available": True,
                "initValue": "str",
                "contentType": "string",
                "protocolType": "String",
                "uid": 10001,
                "name": "Test_Option.2",
            },
            {
                "access": "read",
                "available": True,
                "contentType": "boolean",
                "protocolType": "Boolean",
                "uid": 10002,
                "name": "Test_Option.3",
            },
        ],
        "program": [
            {
                "available": True,
                "uid": 32828,
                "name": "Test_Program",
                "options": [],
            }
        ],
        "activeProgram": {
            "access": "readwrite",
            "uid": 256,
            "name": "BSH.Common.Root.ActiveProgram",
        },
        "selectedProgram": {
            "access": "readwrite",
            "fullOptionSet": False,
            "uid": 257,
            "name": "BSH.Common.Root.SelectedProgram",
        },
    }
)


class MockApplianceType(Protocol):
    """Typeing for mock_homeconnect_appliance fixture."""

    async def __call__(  # noqa: D102
        self,
        description: DeviceDescription = DESCRIPTION,
        host: str = "127.0.0.1",
        app_name: str = TEST_APP_NAME,
        app_id: str = TEST_APP_ID,
        psk64: str = TEST_PSK64,
        iv64: str | None = TEST_IV64,
    ) -> MockAppliance:
        pass



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .entities import (
    ActiveProgram,
    Command,
    DeviceDescription,
    DeviceInfo,
    Entity,
    Event,
    Option,
    Program,
    SelectedProgram,
    Setting,
    Status,
)
from .message import Action, Message
from .session import HCSession

if TYPE_CHECKING:
    from aiohttp import ClientSession


class HomeAppliance:
    """HomeConnect Appliance."""

    session: HCSession
    info: DeviceInfo
    entities_uid: dict[int, Entity]
    "entities by uid"

    entities: dict[str, Entity]
    "entities by name"

    status: dict[str, Status]
    "status entities by name"

    settings: dict[str, Setting]
    "setting entities by name"

    events: dict[str, Event]
    "event entities by name"

    commands: dict[str, Command]
    "command entities by name"

    options: dict[str, Option]
    "option entities by name"

    programs: dict[str, Program]
    "program entities by name"

    _selected_program: SelectedProgram | None = None
    _active_program: ActiveProgram | None = None

    def __init__(  # noqa: PLR0913
        self,
        description: DeviceDescription,
        host: str,
        app_name: str,
        app_id: str,
        psk64: str,
        iv64: str | None = None,
        session: ClientSession | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """
        HomeConnect Appliance.

        Args:
        ----
            description (DeviceDescription): parsed Device description
            host (str): Host
            app_name (str): Name used to identify this App
            app_id (str): ID used to identify this App
            psk64 (str): urlsafe base64 encoded psk key
            iv64 (Optional[str]): urlsafe base64 encoded iv64 key (only AES)
            session (Optional[aiohttp.ClientSession]): ClientSession
            logger (Optional[Logger]): Logger

        """
        if logger is None:
            self._logger = logging.getLogger(__name__)
        else:
            self._logger = logger.getChild("appliance")
        self.session = HCSession(host, app_name, app_id, psk64, iv64, session, logger)
        self.info = description.get("info", {})

        self.entities_uid = {}
        self.entities = {}
        self.status = {}
        self.settings = {}
        self.events = {}
        self.commands = {}
        self.options = {}
        self.programs = {}
        self._create_entities(description)

    async def connect(self) -> None:
        """Open Connection with Appliance."""
        await self.session.connect(self._message_handler)

    async def close(self) -> None:
        """Close Connection with Appliance."""
        await self.session.close()

    async def _message_handler(self, message: Message) -> None:
        """Handel received messages."""
        if message.data is None:
            return
        if message.action == Action.NOTIFY:
            if message.resource in ("/ro/descriptionChange", "/ro/values"):
                await self._update_entities(message.data)
        elif message.action == Action.RESPONSE:
            if message.resource in (
                "/ro/allDescriptionChanges",
                "/ro/allMandatoryValues",
            ):
                await self._update_entities(message.data)
            elif message.resource in ("/iz/info", "/ci/info"):
                # Update device Info
                self.info.update(message.data[0])

    async def _update_entities(self, data: list[dict]) -> None:
        """Update entities from Message data."""
        for entity in data:
            uid = int(entity["uid"])
            if uid in self.entities_uid:
                await self.entities_uid[uid].update(entity)
            else:
                self._logger.debug("Recived Update for unkown entity %s", uid)

    def _create_entities(self, description: DeviceDescription) -> None:
        """Create Entities from Device description."""
        for status in description.get("status", []):
            entity = Status(status, self)
            self.status[entity.name] = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

        for setting in description.get("setting", []):
            entity = Setting(setting, self)
            self.settings[entity.name] = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

        for event in description.get("event", []):
            entity = Event(event, self)
            self.events[entity.name] = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

        for command in description.get("command", []):
            entity = Command(command, self)
            self.commands[entity.name] = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

        for option in description.get("option", []):
            entity = Option(option, self)
            self.options[entity.name] = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

        for program in description.get("program", []):
            entity = Program(program, self)
            self.programs[entity.name] = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

        if "activeProgram" in description:
            entity = ActiveProgram(description["activeProgram"], self)
            self._active_program = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

        if "selectedProgram" in description:
            entity = SelectedProgram(description["selectedProgram"], self)
            self._selected_program = entity
            self.entities[entity.name] = entity
            self.entities_uid[entity.uid] = entity

    async def get_wifi_networks(self) -> list[dict]:
        """Get info on avalibel WiFi networks."""
        msg = Message(resource="/ci/wifiNetworks", action=Action.GET)
        rsp = await self.session.send_sync(msg)
        return rsp.data

    async def get_network_config(self) -> list[dict]:
        """Get current network config."""
        msg = Message(resource="/ni/info", action=Action.GET)
        rsp = await self.session.send_sync(msg)
        return rsp.data

    def dump(self) -> dict:
        """Dump Appliance state."""
        return {
            "entities": [entity.dump() for entity in self.entities.values()],
            "service_versions": self.session.service_versions,
        }

    @property
    def active_program(self) -> Program | None:
        """Return the current Active Program entity or None if no Program is active."""
        return (
            None
            if self._active_program.value == 0 or self._active_program.value is None
            else self.entities_uid[self._active_program.value]
        )

    @property
    def selected_program(self) -> Program | None:
        """Return current selected Program entity or None if no Program is selected."""
        return (
            None
            if self._selected_program.value == 0 or self._selected_program.value is None
            else self.entities_uid[self._selected_program.value]
        )



from __future__ import annotations

import asyncio
import contextlib
import logging
import re
from base64 import urlsafe_b64encode
from json import JSONDecodeError
from typing import TYPE_CHECKING

import aiohttp
from Crypto.Random import get_random_bytes

from .const import (
    DEFAULT_HANDSHAKE_TIMEOUT,
    DEFAULT_SEND_TIMEOUT,
    MAX_CONNECT_TIMEOUT,
    TIMEOUT_INCREASE_FACTOR,
)
from .errors import CodeResponsError, NotConnectedError
from .hc_socket import AesSocket, HCSocket, TlsSocket
from .message import Action, Message, load_message

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


class HCSession:
    """HomeConnect Session."""

    handshake: bool
    """Automatic Handshake"""
    service_versions: dict
    _sid: int | None = None
    _last_msg_id: int | None = None
    _host: str
    _psk64: str
    _iv64: str | None
    _device_info: dict
    _response_messages: dict[int, Message]
    _response_events: dict[int, asyncio.Event]
    _send_lock: asyncio.Lock
    _socket: HCSocket = None
    _recv_loop_event: asyncio.Event
    _connected: bool = False
    _recv_task: asyncio.Task = None
    _handshake_task: asyncio.Task = None
    _tasks: set[asyncio.Task]
    _ext_message_handler: Callable[[Message], None | Awaitable[None]] | None = None

    def __init__(  # noqa: PLR0913
        self,
        host: str,
        app_name: str,
        app_id: str,
        psk64: str,
        iv64: str | None = None,
        session: aiohttp.ClientSession | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """
        HomeConnect Session.

        Args:
        ----
        host (str): Host.
        app_name (str): Name used to identify this App
        app_id (str): ID used to identify this App
        psk64 (str): urlsafe base64 encoded psk key
        iv64 (Optional[str]): urlsafe base64 encoded iv64 key (only AES)
        session (Optional[aiohttp.ClientSession]): ClientSession
        logger (Optional[Logger]): Logger

        """
        self._host = host
        self._psk64 = psk64
        self._iv64 = iv64
        self._device_info = {
            "deviceType": "Application",
            "deviceName": app_name,
            "deviceID": app_id,
        }

        self._recv_loop_event = asyncio.Event()
        self.handshake = True
        self._response_messages = {}
        self._response_events = {}
        self._response_lock = asyncio.Lock()
        self._send_lock = asyncio.Lock()
        self.service_versions = {}
        self._tasks = set()
        self._retry_count = 0

        if logger is None:
            self._logger = logging.getLogger(__name__)
        else:
            self._logger = logger.getChild("session")

        # create socket
        if self._iv64:
            self._logger.debug("Got iv64, using AES socket")
            self._socket = AesSocket(
                self._host, self._psk64, self._iv64, session, logger
            )
        elif self._psk64:
            self._logger.debug("No iv64, using TLS socket")
            self._socket = TlsSocket(self._host, self._psk64, session, logger)
        else:  # For Testing
            self._logger.warning("Using unencrypted socket")
            self._socket = HCSocket(self._host, session, logger)

    @property
    def connected(self) -> bool:
        """Is connected."""
        if self._socket:
            return self._connected and not self._socket.closed
        return False

    async def connect(
        self,
        message_handler: Callable[[Message], Awaitable[None]],
        timeout: int = DEFAULT_HANDSHAKE_TIMEOUT,  # noqa: ASYNC109
    ) -> None:
        """
        Open Connection with Appliance.

        Args:
        ----
        message_handler (Callable[[Message], Awaitable[None]]): called for each message
        timeout (int): timeout (Default: 60).

        """
        self._logger.info("Connecting to %s", self._host)
        self._ext_message_handler = message_handler
        await self._reset()

        try:
            await self._socket.connect()
            self._recv_task = asyncio.create_task(self._recv_loop())
            self._recv_task.add_done_callback(self._recv_loop_done_callback)
            await asyncio.wait_for(self._recv_loop_event.wait(), timeout)
            if not self._connected:
                # loop event received, but not connected
                if self._recv_task.done():
                    # loop exited
                    self._logger.error("Receive loop exited unexpectedly")
                elif self._handshake_task.done():
                    # loop running, handshake eexited
                    if task_exc := self._handshake_task.exception():
                        self._logger.exception("Handshake Exception", exc_info=task_exc)
                        raise task_exc
                    self._logger.error("Handshake exited unexpectedly")
                await self.close()

        except (aiohttp.ClientConnectionError, aiohttp.ClientConnectorSSLError):
            self._logger.exception("Error connecting to Appliance")
            raise
        except TimeoutError:
            if self._recv_task.cancel():
                with contextlib.suppress(asyncio.CancelledError):
                    await self._recv_task
            if self._handshake_task.cancel():
                with contextlib.suppress(asyncio.CancelledError):
                    await self._handshake_task
            self._logger.error("Connection Error")  # noqa: TRY400
            raise

    async def _reset(self) -> None:
        """Rest connction state."""
        self.service_versions.clear()
        self._recv_loop_event.clear()
        self._connected = False
        self._response_messages.clear()
        # Set all response events
        async with self._response_lock:
            for event in self._response_events.values():
                event.set()
            self._response_events.clear()

    async def _recv_loop(self) -> None:
        while self._socket:
            try:
                if self._socket.closed:
                    self._logger.debug(
                        "Socket closed with code %s, opening",
                        self._socket._websocket.close_code,  # noqa: SLF001
                        exc_info=self._socket._websocket.exception(),  # noqa: SLF001
                    )
                    await self._reset()
                    await self._socket.connect()
                async for message in self._socket:
                    # recv messages
                    message_obj = load_message(message)
                    await self._message_handler(message_obj)
            except (aiohttp.ClientConnectionError, aiohttp.ServerTimeoutError) as ex:
                self._logger.warning(ex)
                timeout = TIMEOUT_INCREASE_FACTOR**self._retry_count
                self._retry_count += 1
                await asyncio.sleep(min(timeout, MAX_CONNECT_TIMEOUT))
            except (JSONDecodeError, KeyError):
                self._logger.warning("Can't decode message: %s", message)
            except asyncio.CancelledError:
                self._logger.debug("Receive loop cancelled")
                raise
            except Exception:
                self._logger.exception("Receive loop Exception")

    async def _message_handler(self, message: Message) -> None:
        """Handle recived message."""
        if message.resource == "/ei/initialValues":
            # connection reset/reconncted
            if self._recv_loop_event.is_set():
                self._logger.info("Got init message while connected, resetting")
                await self._reset()
            # set new sID, msgID
            self._sid = message.sid
            self._last_msg_id = message.data[0]["edMsgID"]
            if self.handshake:
                # start handshake
                self._logger.info("Got init message, beginning handshake")
                self._handshake_task = asyncio.create_task(self._handshake(message))
                self._handshake_task.add_done_callback(self._recv_loop_done_callback)
            else:
                self._logger.info("Connected, no handshake")
                self._connected = True
                self._recv_loop_event.set()
                self._retry_count = 0
                await self._call_ext_message_handler(message)

        elif message.action == Action.RESPONSE:
            try:
                async with self._response_lock:
                    if self._response_events[message.msg_id].is_set():
                        # should never happen
                        self._logger.warning(
                            "Response for Msg ID %s was received more then once",
                            message.msg_id,
                        )
                    else:
                        self._response_messages[message.msg_id] = message
                        self._response_events[message.msg_id].set()
            except KeyError:
                self._logger.debug(
                    "Received response for unkown Msg ID %s", message.msg_id
                )

        else:
            # call external message handler
            await self._call_ext_message_handler(message)

    async def _call_ext_message_handler(self, message: Message) -> None:
        """Call the external message handler."""
        task = asyncio.create_task(self._ext_message_handler(message))
        self._tasks.add(task)
        task.add_done_callback(self._done_callback)

    def _done_callback(self, task: asyncio.Task) -> None:
        if exc := task.exception():
            self._logger.exception("Exception in Session callback", exc_info=exc)
        self._tasks.discard(task)

    def _recv_loop_done_callback(self, _: asyncio.Task) -> None:
        self._recv_loop_event.set()

    async def _handshake(self, message_init: Message) -> None:
        try:
            # responde to init message
            await self.send(message_init.responde(self._device_info))

            # request available services
            message_services = Message(resource="/ci/services", version=1)
            response_services = await self.send_sync(message_services)
            self.set_service_versions(response_services)
            await self._call_ext_message_handler(response_services)

            token = urlsafe_b64encode(get_random_bytes(32)).decode("UTF-8")
            token = re.sub(r"=", "", token)
            message_authentication = Message(
                resource="/ci/authentication", data={"nonce": token}
            )
            await self.send(message_authentication)

            # request device info
            with contextlib.suppress(CodeResponsError):
                message_info = Message(resource="/ci/info")
                response_info = await self.send_sync(message_info)
                await self._call_ext_message_handler(response_info)

            with contextlib.suppress(CodeResponsError):
                message_info = Message(resource="/iz/info")
                response_info = await self.send_sync(message_info)
                await self._call_ext_message_handler(response_info)

            # report device ready
            message_ready = Message(resource="/ei/deviceReady", action=Action.NOTIFY)
            await self.send(message_ready)

            message_ready = Message(resource="/ni/info")
            await self.send(message_ready)

            # request mandatory values
            message_mandatory_values = Message(resource="/ro/allMandatoryValues")
            response_mandatory_values = await self.send_sync(message_mandatory_values)
            await self._call_ext_message_handler(response_mandatory_values)

            # request values
            with contextlib.suppress(CodeResponsError):
                message_values = Message(resource="/ro/values")
                response_values = await self.send_sync(message_values)
                await self._call_ext_message_handler(response_values)

            # request description changes
            message_description_changes = Message(resource="/ro/allDescriptionChanges")
            response_description_changes = await self.send_sync(
                message_description_changes
            )
            await self._call_ext_message_handler(response_description_changes)

            # handshake completed
            self._connected = True
            self._recv_loop_event.set()
            self._retry_count = 0
            self._logger.info("Handshake completed")
        except asyncio.CancelledError:
            self._logger.exception("Handshake cancelled")
            raise
        except CodeResponsError:
            self._logger.exception("Received Code response during Handshake")
            raise
        except Exception:
            self._logger.exception("Unknown Exception during Handshake")
            raise

    async def close(self) -> None:
        """Close connction."""
        self._logger.info("Closing connection to %s", self._host)
        if self._recv_task:
            self._recv_task.cancel()
        if self._socket:
            await self._socket.close()
        self._socket = None

    def _set_message_info(self, message: Message) -> None:
        """Set Message infos. called before sending message."""
        # Set service version
        if message.version is None:
            service = message.resource[1:3]
            message.version = self.service_versions.get(service, 1)

        # Set sID
        if message.sid is None:
            message.sid = self._sid

        # Set msgID
        if message.msg_id is None:
            message.msg_id = self._last_msg_id
            self._last_msg_id += 1

    def set_service_versions(self, message: Message) -> None:
        """Set service versions from a '/ci/services' Response."""
        self._logger.debug("Setting Service versions")
        if message.data is not None:
            for service in message.data:
                self.service_versions[service["service"]] = service["version"]
        else:
            msg = "No Data in Message"
            raise ValueError(msg)

    async def send_sync(
        self,
        send_message: Message,
        timeout: float = DEFAULT_SEND_TIMEOUT,  # noqa: ASYNC109
    ) -> Message | None:
        """Send message to Appliance, returns Response Message."""
        response_message: Message | None = None

        async with self._send_lock:
            self._set_message_info(send_message)

            response_event = asyncio.Event()
            async with self._response_lock:
                self._response_events[send_message.msg_id] = response_event

            # send message
            await self._socket.send(send_message.dump())

        try:
            await asyncio.wait_for(response_event.wait(), timeout)
            response_message = self._response_messages[send_message.msg_id]
        except TimeoutError:
            self._logger.debug("Timeout for message %s", send_message.msg_id)
            raise
        except KeyError:
            if not self._connected:
                raise NotConnectedError from None
        finally:
            async with self._response_lock:
                with contextlib.suppress(KeyError):
                    self._response_events.pop(send_message.msg_id)

        if response_message.code:
            self._logger.debug(
                "Received Code %s for Message %s, resource: %s",
                response_message.code,
                send_message.msg_id,
                response_message.resource,
            )
            raise CodeResponsError(response_message.code, response_message.resource)
        return response_message

    async def send(self, message: Message) -> None:
        """Send message to Appliance, returns immediately."""
        async with self._send_lock:
            self._set_message_info(message)
            await self._socket.send(message.dump())



from __future__ import annotations

from typing import ClassVar
from unittest.mock import AsyncMock

import pytest
from homeconnect_websocket.entities import EntityDescription, Event


class MockAppliance(AsyncMock):
    """Mock Appliance."""

    commands: ClassVar = {
        "BSH.Common.Command.AcknowledgeEvent": AsyncMock(),
        "BSH.Common.Command.RejectEvent": AsyncMock(),
    }


@pytest.mark.asyncio
async def test_acknowledge() -> None:
    """Test Event.acknowledge()."""
    description = EntityDescription(
        uid=1,
        name="Test_Event",
        protocolType="Integer",
    )
    appliance = MockAppliance()
    entity = Event(description, appliance)
    await entity.acknowledge()

    appliance.commands[
        "BSH.Common.Command.AcknowledgeEvent"
    ].execute.assert_called_once_with(1)


@pytest.mark.asyncio
async def test_reject() -> None:
    """Test Event.reject()."""
    description = EntityDescription(
        uid=1,
        name="Test_Event",
        protocolType="Integer",
    )
    appliance = MockAppliance()
    entity = Event(description, appliance)
    await entity.reject()

    appliance.commands[
        "BSH.Common.Command.RejectEvent"
    ].execute.assert_called_once_with(1)



from __future__ import annotations

from unittest.mock import AsyncMock

import pytest
from homeconnect_websocket.entities import Access, Command, EntityDescription
from homeconnect_websocket.errors import AccessError
from homeconnect_websocket.message import Action, Message


@pytest.mark.asyncio
async def test_execute() -> None:
    """Test Command.execute()."""
    description = EntityDescription(
        uid=1,
        name="Test_Command",
        access=Access.READ_WRITE,
        available=True,
        protocolType="Integer",
    )
    appliance = AsyncMock()
    entity = Command(description, appliance)
    await entity.execute(5)

    appliance.session.send_sync.assert_called_once_with(
        Message(
            resource="/ro/values",
            action=Action.POST,
            data={"uid": 1, "value": 5},
        )
    )

    entity._available = False
    with pytest.raises(AccessError):
        await entity.execute(5)

    entity._available = True
    entity._access = Access.READ
    with pytest.raises(AccessError):
        await entity.execute(5)



from __future__ import annotations

import json
from pathlib import Path

import pytest
from homeconnect_websocket.description_parser import (
    convert_bool,
    parse_device_description,
)

REFERENCE_DISCRIPTION = {
    "info": {
        "brand": "Fake_Brand",
        "type": "HomeAppliance",
        "model": "Fake_Model",
        "version": 2,
        "revision": 0,
    },
    "status": [
        {
            "uid": 4097,
            "name": "Status.1",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "read",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4098,
            "name": "Status.2",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Open", "1": "Closed"},
            "available": True,
            "access": "read",
            "refCID": 3,
            "refDID": 128,
        },
        {
            "uid": 4099,
            "name": "Status.3",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "read",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4100,
            "name": "Status.4",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Open", "1": "Closed"},
            "available": True,
            "access": "read",
            "refCID": 3,
            "refDID": 128,
        },
    ],
    "setting": [
        {
            "uid": 4101,
            "name": "Setting.1",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "readwrite",
            "min": "0",
            "max": "10",
            "stepSize": "1",
            "initValue": "1",
            "default": "0",
            "passwordProtected": False,
            "notifyOnChange": False,
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4102,
            "name": "Setting.2",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "readwrite",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4103,
            "name": "Setting.3",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "readwrite",
            "min": "0",
            "max": "10",
            "stepSize": "1",
            "initValue": "1",
            "default": "0",
            "passwordProtected": False,
            "notifyOnChange": False,
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4104,
            "name": "Setting.4",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "readwrite",
            "refCID": 1,
            "refDID": 0,
        },
    ],
    "event": [
        {
            "uid": 4105,
            "name": "Event.1",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
            "handling": "acknowledge",
            "level": "hint",
            "refCID": 3,
            "refDID": 128,
        },
        {
            "uid": 4106,
            "name": "Event.2",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
            "handling": "acknowledge",
            "level": "hint",
            "refCID": 3,
            "refDID": 128,
        },
        {
            "uid": 4107,
            "name": "Event.3",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
            "handling": "acknowledge",
            "level": "hint",
            "refCID": 3,
            "refDID": 128,
        },
        {
            "uid": 4108,
            "name": "Event.4",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
            "handling": "acknowledge",
            "level": "hint",
            "refCID": 3,
            "refDID": 128,
        },
    ],
    "command": [
        {
            "uid": 4109,
            "name": "Command.1",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "writeonly",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4110,
            "name": "Command.2",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "writeonly",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4111,
            "name": "Command.3",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "writeonly",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4112,
            "name": "Command.4",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "writeonly",
            "refCID": 1,
            "refDID": 0,
        },
    ],
    "option": [
        {
            "uid": 4113,
            "name": "Option.1",
            "contentType": "percent",
            "protocolType": "Float",
            "available": True,
            "access": "read",
            "liveUpdate": True,
            "refCID": 17,
            "refDID": 160,
        },
        {
            "uid": 4114,
            "name": "Option.2",
            "contentType": "timeSpan",
            "protocolType": "Integer",
            "available": True,
            "access": "read",
            "refCID": 16,
            "refDID": 130,
        },
        {
            "uid": 4115,
            "name": "Option.3",
            "contentType": "percent",
            "protocolType": "Float",
            "available": True,
            "access": "read",
            "liveUpdate": True,
            "refCID": 17,
            "refDID": 160,
        },
        {
            "uid": 4116,
            "name": "Option.4",
            "contentType": "timeSpan",
            "protocolType": "Integer",
            "available": True,
            "access": "read",
            "refCID": 16,
            "refDID": 130,
        },
    ],
    "program": [
        {
            "uid": 4117,
            "name": "Program.1",
            "available": True,
            "execution": "selectonly",
            "options": [
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": False,
                    "refUID": 4113,
                    "default": "true",
                },
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": True,
                    "refUID": 4114,
                },
            ],
        },
        {
            "uid": 4118,
            "name": "Program.2",
            "available": True,
            "options": [
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": False,
                    "refUID": 4113,
                    "default": "true",
                },
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": True,
                    "refUID": 4114,
                },
            ],
        },
        {
            "uid": 4119,
            "name": "Program.3",
            "available": True,
            "execution": "selectonly",
            "options": [
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": False,
                    "refUID": 4113,
                    "default": "true",
                },
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": True,
                    "refUID": 4114,
                },
            ],
        },
        {
            "uid": 4120,
            "name": "Program.4",
            "available": True,
            "options": [
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": False,
                    "refUID": 4113,
                    "default": "true",
                },
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": True,
                    "refUID": 4114,
                },
            ],
        },
    ],
    "activeProgram": {
        "uid": 4121,
        "name": "ActiveProgram",
        "access": "readwrite",
        "validate": True,
    },
    "selectedProgram": {
        "uid": 4122,
        "name": "SelectedProgram",
        "access": "readwrite",
        "fullOptionSet": False,
    },
    "protectionPort": {
        "uid": 4123,
        "name": "ProtectionPort",
        "access": "readwrite",
        "available": True,
    },
}

REFERENCE_DISCRIPTION_SHORT = {
    "info": {
        "brand": "Fake_Brand",
        "type": "HomeAppliance",
        "model": "Fake_Model",
        "version": 2,
        "revision": 0,
    },
    "status": [
        {
            "uid": 4097,
            "name": "Status.1",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "read",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4098,
            "name": "Status.2",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Open", "1": "Closed"},
            "available": True,
            "access": "read",
            "refCID": 3,
            "refDID": 0,
        },
    ],
    "setting": [
        {
            "uid": 4101,
            "name": "Setting.1",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "readwrite",
            "min": "0",
            "max": "10",
            "stepSize": "1",
            "initValue": "1",
            "default": "0",
            "passwordProtected": False,
            "notifyOnChange": False,
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4102,
            "name": "Setting.2",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "readwrite",
            "refCID": 1,
            "refDID": 0,
        },
    ],
    "event": [
        {
            "uid": 4105,
            "name": "Event.1",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
            "handling": "acknowledge",
            "level": "hint",
            "refCID": 3,
            "refDID": 128,
        },
        {
            "uid": 4106,
            "name": "Event.2",
            "contentType": "enumeration",
            "protocolType": "Integer",
            "enumeration": {"0": "Off", "1": "Present", "2": "Confirmed"},
            "handling": "acknowledge",
            "level": "hint",
            "refCID": 3,
            "refDID": 128,
        },
    ],
    "command": [
        {
            "uid": 4109,
            "name": "Command.1",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "writeonly",
            "refCID": 1,
            "refDID": 0,
        },
        {
            "uid": 4110,
            "name": "Command.2",
            "contentType": "boolean",
            "protocolType": "Boolean",
            "available": True,
            "access": "writeonly",
            "refCID": 1,
            "refDID": 0,
        },
    ],
    "option": [
        {
            "uid": 4113,
            "name": "Option.1",
            "contentType": "percent",
            "protocolType": "Float",
            "available": True,
            "access": "read",
            "liveUpdate": True,
            "refCID": 17,
            "refDID": 160,
        },
        {
            "uid": 4114,
            "name": "Option.2",
            "contentType": "timeSpan",
            "protocolType": "Integer",
            "available": True,
            "access": "read",
            "refCID": 16,
            "refDID": 160,
        },
    ],
    "program": [
        {
            "uid": 4117,
            "name": "Program.1",
            "available": True,
            "execution": "selectonly",
            "options": [
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": False,
                    "refUID": 4113,
                    "default": "true",
                },
            ],
        },
        {
            "uid": 4118,
            "name": "Program.2",
            "available": True,
            "execution": "selectonly",
            "options": [
                {
                    "access": "readwrite",
                    "available": True,
                    "liveUpdate": False,
                    "refUID": 4113,
                    "default": "true",
                },
            ],
        },
    ],
    "activeProgram": {
        "uid": 4121,
        "name": "ActiveProgram",
        "access": "readwrite",
        "validate": True,
    },
    "selectedProgram": {
        "uid": 4122,
        "name": "SelectedProgram",
        "access": "readwrite",
        "fullOptionSet": False,
    },
    "protectionPort": {
        "uid": 4123,
        "name": "ProtectionPort",
        "access": "readwrite",
        "available": True,
    },
}


def test_convert_bool() -> None:
    """Test str to bool."""
    assert convert_bool("True") is True
    assert convert_bool("true") is True
    assert convert_bool("False") is False
    assert convert_bool("false") is False

    assert convert_bool(True) is True  # noqa: FBT003
    assert convert_bool(False) is False  # noqa: FBT003

    assert convert_bool(0) is False
    assert convert_bool(1) is True
    assert convert_bool(1.5) is True

    assert convert_bool("0") is False
    assert convert_bool("1") is True
    assert convert_bool("1.5") is True

    with pytest.raises(TypeError):
        convert_bool({})
    with pytest.raises(TypeError):
        convert_bool("not bool")


@pytest.mark.parametrize(
    ("description_path", "feature_path", "expected"),
    [
        (
            Path("tests/DeviceDescription.xml"),
            Path("tests/FeatureMapping.xml"),
            REFERENCE_DISCRIPTION,
        ),
        (
            Path("tests/DeviceDescription_short.xml"),
            Path("tests/FeatureMapping_short.xml"),
            REFERENCE_DISCRIPTION_SHORT,
        ),
    ],
)
def test_parse_device_description(
    description_path: Path, feature_path: Path, expected: dict
) -> None:
    """Test Description Parser."""
    with description_path.open() as file:
        description_file = file.read()
    with feature_path.open() as file:
        feature_file = file.read()

    paresd_description = parse_device_description(description_file, feature_file)
    paresd_description = json.loads(json.dumps(paresd_description))

    assert paresd_description == expected



from .appliance import HomeAppliance
from .description_parser import parse_device_description
from .entities import DeviceDescription
from .errors import (
    AccessError,
    CodeResponsError,
    HomeConnectError,
    NotConnectedError,
    ParserError,
)
from .message import Message

__all__ = [
    "AccessError",
    "CodeResponsError",
    "DeviceDescription",
    "HomeAppliance",
    "HomeConnectError",
    "Message",
    "NotConnectedError",
    "ParserError",
    "parse_device_description",
]



from __future__ import annotations

import json
import re
from dataclasses import dataclass
from enum import StrEnum


class Action(StrEnum):
    """Message Actions."""

    GET = "GET"
    POST = "POST"
    RESPONSE = "RESPONSE"
    NOTIFY = "NOTIFY"


def load_message(msg_str: str | dict) -> Message:
    """Load Message from json."""
    message = Message()
    msg = json.loads(msg_str) if isinstance(msg_str, str) else msg_str
    message.sid = int(msg["sID"])
    message.msg_id = int(msg["msgID"])
    message.resource = msg["resource"]
    message.version = int(msg["version"])
    message.action = Action(msg["action"])
    message.data = msg.get("data", None)
    message.code = msg.get("code", None)
    return message


@dataclass
class Message:
    """Represents an Websocket Message."""

    sid: int | None = None
    """Session ID"""
    msg_id: int | None = None
    """Message ID"""
    resource: str | None = None
    """Resource Endpoint"""
    version: int | None = None
    """Service Version"""
    action: Action = Action.GET
    """Action"""
    data: list[dict] | None = None
    """Message Data"""
    code: int | None = None
    """Response Code"""

    def responde(self, data: list[dict] | None = None) -> Message:
        """Generate a response Message."""
        return Message(
            sid=self.sid,
            msg_id=self.msg_id,
            resource=self.resource,
            version=self.version,
            action=Action.RESPONSE,
            data=data,
        )

    def dump(self) -> str:
        """Dump message to string."""
        msg = {
            "sID": self.sid,
            "msgID": self.msg_id,
            "resource": self.resource,
            "version": self.version,
            "action": self.action.value,
        }
        if self.data is not None:
            # data must be list
            if isinstance(self.data, list):
                msg["data"] = self.data
            else:
                msg["data"] = [self.data]
        buf = json.dumps(msg, separators=(",", ":"))
        # swap ' for ""
        return re.sub("'", '"', buf)



from __future__ import annotations

from homeconnect_websocket.message import Action, Message, load_message


def test_message_load() -> None:
    """Test load_message()."""
    message_str = '{"sID":3609751529,"msgID":3294751923,"resource":"/ei/initialValues","version":2,"action":"POST","data":[{"edMsgID":2974143846}]}'  # noqa: E501

    message = load_message(message_str)
    assert message.sid == 3609751529
    assert message.msg_id == 3294751923
    assert message.resource == "/ei/initialValues"
    assert message.version == 2
    assert message.action == Action.POST
    assert message.data == [{"edMsgID": 2974143846}]
    assert message.code is None

    message_str = '{"sID":3609751529,"msgID":3294751923,"resource":"/ro/values","version":1,"action":"RESPONSE","code":400}'  # noqa: E501
    message = load_message(message_str)
    assert message.resource == "/ro/values"
    assert message.version == 1
    assert message.action == Action.RESPONSE
    assert message.data is None
    assert message.code == 400


def test_message_dump() -> None:
    """Test Message.dump()."""
    message = Message(
        sid=3609751529,
        msg_id=3294751923,
        resource="/ei/initialValues",
        version=2,
        action=Action.POST,
        data={"edMsgID": 2974143846},
    )
    assert (
        message.dump()
        == '{"sID":3609751529,"msgID":3294751923,"resource":"/ei/initialValues","version":2,"action":"POST","data":[{"edMsgID":2974143846}]}'  # noqa: E501
    )


def test_message_response() -> None:
    """Test Message.responde()."""
    message = Message(
        sid=3609751529,
        msg_id=3294751923,
        resource="/ei/initialValues",
        version=2,
        action=Action.POST,
        data={"edMsgID": 2974143846},
    )
    response_message = message.responde({"deviceType": "Application"})

    assert response_message.sid == 3609751529
    assert response_message.msg_id == 3294751923
    assert response_message.resource == "/ei/initialValues"
    assert response_message.version == 2
    assert response_message.action == Action.RESPONSE
    assert response_message.data == {"deviceType": "Application"}
    assert response_message.code is None



"""Helper functions."""

from __future__ import annotations

import contextlib
import json
from typing import Final


def convert_bool(obj: str | bool | float) -> bool:
    """Convert a string to as bool."""
    if isinstance(obj, str):
        if obj.lower() == "true":
            return True
        if obj.lower() == "false":
            return False
        with contextlib.suppress(ValueError):
            obj = float(obj)
    if isinstance(obj, bool):
        return obj
    if isinstance(obj, float | int):
        return bool(obj)
    msg = "Can't convert %s to bool"
    raise TypeError(msg, obj)


def load_object(obj: str) -> dict:
    """Load complex objects from json strings."""
    if isinstance(obj, str):
        return json.loads(obj)
    return obj


TYPE_MAPPING: Final[dict[str, type]] = {
    "Boolean": convert_bool,
    "Integer": int,
    "Float": float,
    "String": str,
    "Object": load_object,
    None: lambda value: value,
}



from __future__ import annotations


class HomeConnectError(Exception):
    """General HomeConnect exception."""


class CodeResponsError(HomeConnectError):
    """Code Response Recived from Appliance."""

    def __init__(self, code: int, resource: str, *args: object) -> None:
        """
        Code Response Recived from Appliance.

        Args:
        ----
        code (int): Recived Code
        resource (str): Recived resource
        *args (object): extra args

        """
        self.code = code
        self.resource = resource
        super().__init__(*args)

    def __str__(self) -> str:
        return f"{self.code}, resource={self.resource}"


class AccessError(HomeConnectError):
    """Entity not Accessible."""


class NotConnectedError(HomeConnectError):
    """Client is not Connected."""


class ParserError(HomeConnectError):
    """Description Parser Error."""



from __future__ import annotations

import asyncio
import logging
from abc import ABC
from enum import StrEnum
from typing import TYPE_CHECKING, Any, TypedDict

from .errors import AccessError
from .helpers import TYPE_MAPPING
from .message import Action, Message

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

    from .appliance import HomeAppliance

_LOGGER = logging.getLogger(__name__)


class Access(StrEnum):
    """Access levels."""

    NONE = "none"
    READ = "read"
    READ_WRITE = "readwrite"
    WRITE_ONLY = "writeonly"
    READ_STATIC = "readstatic"


class EventLevel(StrEnum):
    """Event Levels."""

    INFO = "info"
    HINT = "hint"
    WARNING = "warning"
    ALERT = "alert"
    CRITOCAL = "critical"


class EventHandling(StrEnum):
    """Event handling types."""

    NONE = "none"
    ACKNOWLEDGE = "acknowledge"
    DECISION = "decision"


class Execution(StrEnum):
    """Execution types."""

    NONE = "none"
    SELECT_ONLY = "selectonly"
    START_ONLY = "startonly"
    SELECT_AND_START = "selectandstart"


class DeviceInfo(TypedDict, total=False):
    """Typing for Device info."""

    brand: str
    type: str
    model: str
    version: int
    revision: int
    deviceID: str
    eNumber: str
    vib: str
    mac: str
    haVersion: str
    swVersion: str
    hwVersion: str
    deviceType: str
    deviceInfo: str
    customerIndex: str
    serialNumber: str
    fdString: str
    shipSki: str


class OptionDescription(TypedDict, total=False):
    """Typing for Option Description."""

    access: Access
    available: bool
    liveUpdate: bool
    refUID: int
    default: Any


class EntityDescription(TypedDict, total=False):
    """Typing for Entity Description."""

    uid: int
    name: str
    enumeration: dict | None
    available: bool
    access: Access
    min: int | float
    max: int | float
    stepSize: int | float
    notifyOnChange: bool
    initValue: Any
    passwordProtected: bool
    handling: EventHandling
    level: EventLevel
    default: Any
    liveUpdate: bool
    refUID: int
    options: list[OptionDescription]
    execution: Execution
    fullOptionSet: bool
    validate: bool
    refCID: int
    refDID: int
    protocolType: str
    contentType: str


class DeviceDescription(TypedDict, total=False):
    """Typing for DeviceDescription."""

    info: DeviceInfo
    status: list[EntityDescription]
    setting: list[EntityDescription]
    event: list[EntityDescription]
    command: list[EntityDescription]
    option: list[EntityDescription]
    program: list[EntityDescription]
    activeProgram: EntityDescription
    selectedProgram: EntityDescription
    protectionPort: EntityDescription


class Entity(ABC):
    """BaseEntity Class."""

    _appliance: HomeAppliance
    _uid: int
    _name: str
    _callbacks: set[Callable[[Entity], Coroutine]]
    _value: Any | None = None
    _enumeration: dict | None = None
    _rev_enumeration: dict

    def __init__(
        self, description: EntityDescription, appliance: HomeAppliance
    ) -> None:
        """BaseEntity Class."""
        self._appliance: HomeAppliance = appliance
        self._uid = description["uid"]
        self._name = description["name"]
        self._callbacks = set()
        self._tasks = set()
        self._type = TYPE_MAPPING.get(
            description.get("protocolType"), lambda value: value
        )
        if "enumeration" in description:
            self._enumeration = {
                int(k): v for k, v in description["enumeration"].items()
            }
            self._rev_enumeration = {
                v: int(k) for k, v in description["enumeration"].items()
            }
        if "initValue" in description:
            self._value = self._type(description["initValue"])
        if "default" in description:
            self._value = self._type(description["default"])

    async def update(self, values: dict) -> None:
        """Update the entity state and execute callbacks."""
        if "value" in values:
            self._value = self._type(values["value"])

        for callback in self._callbacks:
            try:
                task = asyncio.create_task(callback(self))
                self._tasks.add(task)
                task.add_done_callback(self._done_callback)
            except Exception:
                _LOGGER.exception("Callback for %s raised an Exception", self.name)

    def _done_callback(self, task: asyncio.Task) -> None:
        if exc := task.exception():
            _LOGGER.exception(
                "Exception in callback for entity %s", self.name, exc_info=exc
            )
        self._tasks.discard(task)

    def register_callback(self, callback: Callable[[Entity], Coroutine]) -> None:
        """Register update callback."""
        if callback not in self._callbacks:
            self._callbacks.add(callback)

    def unregister_callback(self, callback: Callable[[Entity], Coroutine]) -> None:
        """Unregister update callback."""
        self._callbacks.remove(callback)

    def dump(self) -> dict:
        """Dump Entity state."""
        return {
            "uid": self.uid,
            "name": self.name,
            "value": self.value,
            "value_raw": self.value_raw,
            "enum": self.enum,
        }

    @property
    def uid(self) -> int:
        """Entity uid."""
        return self._uid

    @property
    def name(self) -> str:
        """Entity name."""
        return self._name

    @property
    def value(self) -> Any | None:
        """
        Current Value of the Entity.

        if the Entity is an Enum entity the value will be resolve to the actual value.
        """
        if self._enumeration and self._value is not None:
            return self._enumeration[self._value]
        return self._value

    async def set_value(self, value: str | int | bool) -> None:
        """
        Set the Value of the Entity.

        if the Entity is an Enum entity the value will be resolve to the reference Value
        """
        if self._enumeration:
            if value not in self._rev_enumeration:
                msg = "Value not in Enum"
                raise ValueError(msg)
            await self.set_value_raw(self._rev_enumeration[value])
        else:
            await self.set_value_raw(value)

    @property
    def value_raw(self) -> Any | None:
        """Current raw Value."""
        return self._value

    async def set_value_raw(self, value_raw: str | float | bool) -> None:
        """Set the raw Value."""
        message = Message(
            resource="/ro/values",
            action=Action.POST,
            data={"uid": self._uid, "value": self._type(value_raw)},
        )
        await self._appliance.session.send_sync(message)

    @property
    def enum(self) -> dict[int, str] | None:
        """The internal enumeration."""
        return self._enumeration


class AccessMixin(Entity):
    """Mixin for Entities with access attribute."""

    _access: Access | None = None

    def __init__(
        self, description: EntityDescription, appliance: HomeAppliance
    ) -> None:
        """
        Mixin for Entities with access attribute.

        Args:
        ----
            description (EntityDescription): The entity description
            appliance (HomeAppliance): Appliance

        """
        super().__init__(description, appliance)
        self._access = description.get("access", self._access)

    async def update(self, values: dict) -> None:
        """Update the entity state and execute callbacks."""
        if "access" in values:
            self._access = Access(values["access"].lower())
        await super().update(values)

    @property
    def access(self) -> Access | None:
        """Current Access state."""
        return self._access

    async def set_value_raw(self, value_raw: str | int | bool) -> None:
        """Set the raw Value."""
        if self._access not in [Access.READ_WRITE, Access.WRITE_ONLY]:
            msg = "Not Writable"
            raise AccessError(msg)
        await super().set_value_raw(value_raw)

    def dump(self) -> dict:
        """Dump Entity state."""
        state = super().dump()
        state["access"] = self.access
        return state


class AvailableMixin(Entity):
    """Mixin for Entities with available attribute."""

    _available: bool | None = None

    def __init__(
        self, description: EntityDescription, appliance: HomeAppliance
    ) -> None:
        """
        Mixin for Entities with available attribute.

        Args:
        ----
            description (EntityDescription): The entity description
            appliance (HomeAppliance): Appliance

        """
        super().__init__(description, appliance)
        self._available = description.get("available", self._available)

    async def update(self, values: dict) -> None:
        """Update the entity state and execute callbacks."""
        if "available" in values:
            self._available = bool(values["available"])
        await super().update(values)

    @property
    def available(self) -> bool | None:
        """Current Available state."""
        return self._available

    async def set_value_raw(self, value_raw: str | int | bool) -> None:
        """Set the raw Value."""
        if not self._available:
            msg = "Not Available"
            raise AccessError(msg)
        await super().set_value_raw(value_raw)

    def dump(self) -> dict:
        """Dump Entity state."""
        state = super().dump()
        state["available"] = self.available
        return state


class MinMaxMixin(Entity):
    """Mixin for Entities with available Min and Max values."""

    _min: float | None = None
    _max: float | None = None
    _step: float | None = None

    def __init__(
        self, description: EntityDescription, appliance: HomeAppliance
    ) -> None:
        """
        Mixin for Entities with available Min and Max values.

        Args:
        ----
            description (EntityDescription): The entity description
            appliance (HomeAppliance): Appliance

        """
        super().__init__(description, appliance)
        if "min" in description:
            self._min = float(description["min"])
        if "max" in description:
            self._max = float(description["max"])
        if "stepSize" in description:
            self._step = float(description["stepSize"])

    @property
    def min(self) -> float | None:
        """Minimum value."""
        return self._min

    @property
    def max(self) -> float | None:
        """Maximum value."""
        return self._max

    @property
    def step(self) -> float | None:
        """Minimum value."""
        return self._step

    def dump(self) -> dict:
        """Dump Entity state."""
        state = super().dump()
        state["min"] = self.min
        state["max"] = self.max
        state["step"] = self.step
        return state


class Status(AccessMixin, AvailableMixin, MinMaxMixin, Entity):
    """Represents an Settings Entity."""


class Setting(AccessMixin, AvailableMixin, MinMaxMixin, Entity):
    """Represents an Settings Entity."""


class Event(Entity):
    """Represents an Event Entity."""

    async def acknowledge(self) -> None:
        """Acknowledge Event."""
        await self._appliance.commands["BSH.Common.Command.AcknowledgeEvent"].execute(
            self._uid
        )

    async def reject(self) -> None:
        """Reject Event."""
        await self._appliance.commands["BSH.Common.Command.RejectEvent"].execute(
            self._uid
        )


class Command(AccessMixin, AvailableMixin, MinMaxMixin, Entity):
    """Represents an Command Entity."""

    async def execute(self, value: str | int | bool) -> None:
        """Execute command."""
        if self._access not in [Access.READ_WRITE, Access.WRITE_ONLY]:
            msg = "Not Writable"
            raise AccessError(msg)

        if not self._available:
            msg = "Not Available"
            raise AccessError(msg)

        message = Message(
            resource="/ro/values",
            action=Action.POST,
            data={"uid": self._uid, "value": value},
        )
        await self._appliance.session.send_sync(message)


class Option(AccessMixin, AvailableMixin, MinMaxMixin, Entity):
    """Represents an Option Entity."""


class Program(AvailableMixin, Entity):
    """Represents an Program Entity."""

    def __init__(
        self, description: EntityDescription, appliance: HomeAppliance
    ) -> None:
        """
        Program Entity.

        Args:
        ----
            description (EntityDescription): parsed Device description
            appliance (HomeAppliance): Host

        """
        super().__init__(description, appliance)
        self._options: list[Option] = []
        if "options" in description:
            for option in description["options"]:
                self._options.append(appliance.entities_uid[option["refUID"]])

    async def select(self) -> None:
        """Select this Program."""
        message = Message(
            resource="/ro/selectedProgram",
            action=Action.POST,
            data={"program": self._uid, "options": []},
        )
        await self._appliance.session.send_sync(message)

    async def start(self) -> None:
        """Start this Program, select might be required first."""
        options = [
            {"uid": option.uid, "value": option.value_raw}
            for option in self._options
            if option.access == Access.READ_WRITE
        ]
        message = Message(
            resource="/ro/activeProgram",
            action=Action.POST,
            data={"program": self._uid, "options": options},
        )
        await self._appliance.session.send_sync(message)


class ActiveProgram(AccessMixin, AvailableMixin, Entity):
    """Represents the Active_Program Entity."""

    _available = True


class SelectedProgram(AccessMixin, AvailableMixin, Entity):
    """Represents the Selected_Program Entity."""

    _available = True


class ProtectionPort(AccessMixin, AvailableMixin, Entity):
    """Represents an Protection_Port Entity."""

    _available = False
